// test111.cc is a part of the PYTHIA event generator.
// Copyright (C) 2022 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This test program will output PDFs for an extensive number of hadrons,
// as well as check that various symmetry properties are satisfied.

#include "Pythia8/Pythia.h"

using namespace Pythia8;

// Which beams should the PDFs be evaluated for.
vector<int> idBeams = {
  11, 13, 15, 21, 22, 990,
  111, 211, 213, 223, 221, 331,
  311, 321, 313, 130, 310, 333,
  411, 421, 413, 431, 443,
  511, 521, 513, 531, 541, 553,
  2212, 2112, 1114, 2114, 2214, 2224,
  3222, 3212, 3112, 3224, 3122, 3322, 3312, 3334,
  4222, 4212, 4112, 4322, 4312,
  5222, 5212, 5112, 5322, 5312
};

// Which content can appear in the PDFs.
vector<int> qs = {
  1, -1, 2, -2, 3, -3, 4, -4, 5, -5, 21, 22, 11, 13, 15 };

// Where PDFs should be evaluated.
constexpr double x = 0.5, Q2 = 100;

int main() {

  Pythia pythia;
  pythia.readString("ProcessLevel:all = off");
  pythia.init();

  ofstream cmp("test111.dat");

  for (int q : qs) cmp << " " << q;
  cmp << endl;

  for (int idBeam : idBeams) {
    PDFPtr pdf = pythia.getPDFPtr(idBeam);
    PDFPtr antipdf = pythia.getPDFPtr(pythia.particleData.antiId(idBeam));

    if (pdf == nullptr || antipdf == nullptr) {
      cmp << "Got nullptr for " << idBeam << endl;
      continue;
    }

    cmp << "Testing for " << idBeam << endl;

    vector<double> xfs(qs.size()), xvs(qs.size()), xss(qs.size());
    for (size_t iq = 0; iq < qs.size(); ++iq) {
      double q = qs[iq];
      double aq = (q == 22) ? 22 : (q == 21) ? 0 : -q;
      double xf = pdf->xf(q, x, Q2);
      double xv = pdf->xfVal(q, x, Q2);
      double xs = pdf->xfSea(q, x, Q2);
      double af = antipdf->xf(aq, x, Q2);
      double av = antipdf->xfVal(aq, x, Q2);
      double as = antipdf->xfSea(aq, x, Q2);

      // Store value for printing later.
      xfs[iq] = xf;
      xvs[iq] = xv;
      xss[iq] = xs;

      // Particle--antiparticle symmetry
      if (xf != af && (idBeam != 130 && idBeam != 310))
        cmp << " ERROR: xf antiparticle symmetry failed for " << q << endl;
      if (xv != av && (idBeam != 130 && idBeam != 310))
        cmp << " ERROR: Valence antiparticle symmetry failed for " << q << endl;
      if (xs != as)
        cmp << " ERROR: Sea antiparticle symmetry failed for " << q << endl;

      // q and qbar sea content should be equal.
      if ( xs != pdf->xfSea(aq, 0.5, Q2) )
          cmp << "ERROR: Sea is asymmetric for " << q << endl;

      // Ensure q = v + s
      if (abs( xf - (xv + xs) ) > 1e-9)
        cmp << "ERROR: q != v + s for " << q << endl;
    }

    // Print stored values.
    for (double f : xfs) cmp << " " << f;
    cmp << endl;
    for (double f : xvs) cmp << " " << f;
    cmp << endl;
    for (double f : xss) cmp << " " << f;
    cmp << endl;
  }
}
