// test201.cc is a part of the PYTHIA event generator.
// Copyright (C) 2022 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Tests VINCIA shower model, on ee -> ffbar @ 500 GeV, with f = 2nd and 3rd
// generation, including leptons and neutrinos. (Requires --with-mg5mes).
// Otherwise similar to test200, with just some basic multiplicity counts. 

// Include Pythia8 header(s) and namespace.
#include "Pythia8/Pythia.h"
using namespace Pythia8;

// Main Program
int main() {

  //************************************************************************
  // Define Pythia 8 generator

  Pythia pythia;

  // Read user settings from file
  pythia.readFile("test201.cmnd");

  //************************************************************************

  // Shorthands
  Event& event = pythia.event;
  Settings& settings = pythia.settings;

  // Extract settings to be used in the main program.
  int    nEvent     = settings.mode("Main:numberOfEvents");
  int    nAbort     = settings.mode("Main:timesAllowErrors");
  bool   vinciaOn   = settings.mode("PartonShowers:model") == 2;
  bool   helicityOn = vinciaOn && settings.flag("Vincia:helicityShower");
  int    iEventPri  = -1;

  //************************************************************************

  // Initialize
  if(!pythia.init()) { return EXIT_FAILURE; }

  //************************************************************************

  // Counters.
  double nQsum      = 0.0;
  double nPsum      = 0.0;
  double nCsum      = 0.0;
  double nStrangeSum = 0.0;
  double nCharmSum  = 0.0;
  double nBottomSum = 0.0;
  double nGammaSum  = 0.0;
  double nLepSum    = 0.0;
  double nNeutSum   = 0.0;

  //************************************************************************

  // EVENT GENERATION LOOP.
  // Generation, event-by-event printout, analysis, and histogramming.

  // Counter for negative-weight events
  double weight=1.0;
  double sumWeights = 0.0;

  // Begin event loop
  int iAbort = 0;
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    bool aborted = !pythia.next();
    if(aborted){
      event.list();
      if (++iAbort < nAbort){
        continue;
      }
      cout << " Event generation aborted prematurely, owing to error!\n";
      cout<< "Event number was : "<<iEvent<<endl;
      break;
    }

    // Check for weights
    weight = pythia.info.weight();
    sumWeights += weight;

    // Print event with helicities
    if (iEvent == 0 || iEvent == iEventPri)
      if (helicityOn) event.list(helicityOn);

    // Count FS charged hadrons, partons, and quarks
    double nc=0;
    int np=0;
    int nq=0;
    int nGam=0;
    int nLep=0;
    int nNeut=0;
    int nStrange=0;
    int nCharm  =0;
    int nBottom =0;    
    int nLepBorn  = (event[6].isCharged() && event[6].isLepton()) ? 2: 0;
    int nNeutBorn = (!event[6].isCharged() && event[6].isLepton()) ? 2: 0;
    int nUDBorn = (event[6].idAbs() <= 2) ? 2 : 0;
    int nStrangeBorn = (event[6].idAbs() == 3) ? 2 : 0;
    int nCharmBorn   = (event[6].idAbs() == 4) ? 2 : 0;
    int nBottomBorn  = (event[6].idAbs() == 5) ? 2 : 0;
    for (int i=1;i<event.size();i++) {
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) nc++;
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
        // Count up partons and quarks
        if (event[i].isQuark() || event[i].isGluon()) np++;
        if (event[i].id() == 22) nGam++;
        if (event[i].isQuark()) nq++;
        if (event[i].idAbs() == 3) nStrange++;
        if (event[i].idAbs() == 4) nCharm++;
        if (event[i].idAbs() == 5) nBottom++;
        if (event[i].isLepton()) {
          if (event[i].isCharged()) nLep++;
          else nNeut++;
        }
      }
    }
    nQsum += (nq - nUDBorn - nStrangeBorn - nCharmBorn - nBottomBorn) * weight;
    nPsum += np * weight;
    nCsum += nc * weight;
    nStrangeSum += (nStrange - nStrangeBorn) * weight;
    nCharmSum   += (nCharm - nCharmBorn) * weight;
    nBottomSum  += (nBottom - nBottomBorn) * weight;
    nLepSum  += (nLep - nLepBorn) * weight;
    nNeutSum  += (nNeut - nNeutBorn) * weight;
    nGammaSum   += nGam * weight;

  }

  //************************************************************************

  // POST-RUN FINALIZATION
  // Normalization, Statistics, Output.

  //Normalize histograms to effective number of positive-weight events.
  double normFac=1.0/sumWeights;

  // Print out end-of-run information.
  pythia.stat();
  ofstream ofs("test201.dat");
  ofs<<endl;
  ofs<<fixed;
  ofs<< " <nQ - nQBorn>   = "<<num2str(nQsum * normFac / 2.)<<endl;
  ofs<< " <nS - nSBorn>   = "<<num2str(nStrangeSum * normFac / 2.)<<endl;
  ofs<< " <nC - nCBorn>   = "<<num2str(nCharmSum * normFac / 2.)<<endl;
  ofs<< " <nB - nBBorn>   = "<<num2str(nBottomSum * normFac / 2.)<<endl;
  ofs<< " <nL - nLBorn>   = "<<num2str(nLepSum * normFac / 2.)<<endl;
  ofs<< " <nNu - nNuBorn> = "<<num2str(nNeutSum * normFac / 2.)<<endl;
  ofs<< " <n(Q+Qbar+G)> = "<<num2str(nPsum * normFac)<<endl;
  ofs<< " <nGam>        = "<<num2str(nGammaSum * normFac)<<endl;
  ofs<< " <nCharged>    = "<<num2str(nCsum * normFac)<<endl;
  ofs<<endl;

  // Done.
  return 0;
}
