// trimfile.cc is a part of the PYTHIA event generator.
// Copyright (C) 2022 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL v2 or later, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Execute with "private/trimfile filename" from main directory.

// Perform/prepare a number of cleanup operations on a file:
// -- remove unwanted characters at the end of each line
//    (keep one blank for xml files to simplify reformatting);
// -- remove empty lines at the end of the file;
// -- warn if a line is too long
//    (but leave reformatting to manual control);
// -- warn if a line contains tab characters
//    (can be fixed in emacs by "ctrl-x h" followed by "cmnd-x untabify");
// -- perform some string replacements, notably copyright year 
//    (&copy; in Frontpage.xml to be done by hand; also update convertXML).
// -- optionally check and improve formatting, with option "-m"; currently
//    this includes:
//      - warnings if multiple semicolons on a single line (outside
//        of parentheses, quotation marks, or comments). 
//      - check for spaces after commas (outside quotation marks) and insert
//        if missing. 

// By default trimfile only works on .h, .cc, .cmnd and .xml files,
// but command-line option "-f" forces it for any file type.

// With the command-line option "-d" you can give a directory name as input,
// and all the files in it will be processed. Does not handle subdirectories,
// and cannot be combined with "-f". 

//==========================================================================

// Stdlib header files.
#include <vector>
#include <cctype>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <dirent.h>

// Used Stdlib elements.
using std::string;
using std::vector;
using std::cout;
using std::endl;
using std::ifstream;
using std::ofstream;
using std::istringstream;

//--------------------------------------------------------------------------

// Recommended max line length.
int maxLen = 79;

// Possibility to replace strings inside file.
string replaceOld[4] = { "Copyright (C) 2022", "GNU GPL version 2",
  "\"true\"", "\"false\"" }; 
string replaceNew[4] = { "Copyright (C) 2022", "GNU GPL v2 or later",
  "\"on\"",   "\"off\"" };
int    replaceLen[4] = { 18, 17, 6, 7 };

//--------------------------------------------------------------------------

// Forward reference.
bool trimFile(string fileName, bool doPad, bool doSemicolons);

//--------------------------------------------------------------------------

int main(int argc, char* argv[]) {

  // Command-line arguments.
  vector<string> args;
  if (argc > 1) args.push_back(string(argv[1])); 
  if (argc > 2) args.push_back(string(argv[2]));
  if (argc > 3) args.push_back(string(argv[3]));
  if (argc > 4) args.push_back(string(argv[4]));
  if (args.size() == 0) args.push_back("");
  
  bool doForce = false;
  bool doDir   = false;
  bool doFormatting = false;
  int iargFileOrDir = 0;
  for (int iarg = 0; iarg < args.size(); ++iarg) {
    if (args[iarg] == "-f") doForce = true;
    else if (args[iarg] == "-d") doDir = true;
    else if (args[iarg] == "-m") doFormatting = true;
    else iargFileOrDir = iarg;
  }
  
  // Check for incompatible arguments
  if ( doForce && doDir ) {
    cout << "\n Error: options -d and -f cannot be used simultaneously.\n";
    return 1;
  }

  // Vector of files to be read. Only one if not directory.
  vector<string> files, problemFiles;
  string fileOrDir = args[iargFileOrDir];
  if (!doDir) files.push_back( fileOrDir);

  // Open a directory of files. Check that it worked.
  else {
    DIR* directory;
    struct dirent* direntry;
    directory = opendir( fileOrDir.c_str() );
    if (!directory) {
      cout << " Error: failed to open directory " << fileOrDir << endl;
      return 1;
    }

    // Read names of files in directory. Skip hidden files.
    while ( (direntry = readdir(directory)) ) {
      string fileTmp = direntry->d_name;
      if (fileTmp[0] != '.') files.push_back( fileOrDir + "/" + fileTmp);
    }
    closedir(directory);
  }

  // Loop over files in vector.
  for (unsigned int i = 0; i < files.size(); ++i) {
    string fileName = files[i];

    // Identify file type.
    int nameLen = fileName.size();
    bool isCode = (nameLen > 2 && fileName.substr(nameLen - 2, 2) == ".h")
               || (nameLen > 3 && fileName.substr(nameLen - 3, 3) == ".cc")
               || (nameLen > 5 && fileName.substr(nameLen - 5, 5) == ".cmnd");
    bool isXML  = (nameLen > 4 && fileName.substr(nameLen - 4, 4) == ".xml");

    // Skip FJcore and MixMax files; they give plenty of warnings 
    // not to be addressed.
    if (fileName.find("FJcore") != string::npos) isCode = false;
    if (fileName.find("MixMax") != string::npos) isCode = false;

    // Process code and xml files; also others with -f option.
    if (isCode || isXML || doForce) {
      cout << "\n Begin trimming file " << fileName << endl;
      if (!trimFile( fileName, isXML, doFormatting))
        problemFiles.push_back( fileName);
    } else {
      cout << "\n File " << fileName << " is not processed" << endl;
    }
  }

  // Write error summary info for processing of files.
  if (doDir && problemFiles.size() > 0) {
    cout << "\n The following files have unresolved problems: " << endl;
    for (int i = 0; i < int(problemFiles.size()); ++i)
      cout << "    " <<problemFiles[i] << endl; 
  } 

  // Done.
  return 0;
}

//--------------------------------------------------------------------------

bool trimFile(string fileName, bool doPad, bool doFormatting) {

  bool allFine = true;

  // Open input file.
  ifstream is( fileName.c_str() );
  if (!is) {
    cout << " Error: input file " << fileName << " not found" << endl;
    return false;
  }

  // Read in input file.
  vector<string> lines;
  string line;
  while ( getline(is, line) ) lines.push_back(line);
  is.close();

  // How many kinds of replacements? Warn for long lines?
  int  kindRep  = doPad ? 4 : 2;
  bool warnLong = (fileName.find("ParticleData.xml") == string::npos); 

  // Counters for number of trims, removals, replacements, and spaces.
  int nTrim    = 0;
  int nRemove  = lines.size();
  int nReplace = 0;
  int nSpaces  = 0;

  // Loop over all lines in the file.
  for (unsigned int i = 0; i < lines.size(); ++i) {
    line = lines[i];

    // Trim the line to have zero or one blank at the end, but no more.
    size_t posEnd = line.find_last_not_of(" \n\t\v\b\r\f\a");
    if (posEnd == string::npos) posEnd = -1;
    line.erase(posEnd + 1);
    if (doPad) line += " ";
    if (line != lines[i]) ++nTrim;

    // Do string replacements inside line where required.
    for (int j = 0; j < kindRep; ++j)
    if (line.find(replaceOld[j]) != string::npos) {
      line.replace( line.find(replaceOld[j]), replaceLen[j], replaceNew[j]);
      ++nReplace;
    }

    // Count number of semicolons outside parentheses, qoutation marks,
    // or comments.
    if (doFormatting) {
      int nSemicolon = 0;
      int parLevel   = 0;
      bool inQuotes  = false;
      for (int pos = 0; pos < line.size(); ++pos) {
        if (line[pos] == '\"') inQuotes = !inQuotes;
        else if (line[pos] == '(') parLevel++;
        else if (line[pos] == ')') parLevel = std::max( 0, --parLevel);
        if (inQuotes || parLevel >= 1) continue;
        if (line[pos] == ';') nSemicolon++;
      }
      if (nSemicolon >= 2) {
        cout << "    Warning: line " << i + 1
             << " contains multiple semicolons " << endl;    
        allFine = false;
      }

      // Check if commas or equals signs should be followed by space.
      // Also check for spaces around + signs. Minus signs not done, much more
      // difficult due to operators like ->, 1e-16, -4, etc.
      // Skip if in quotes or comments. 
      inQuotes  = false;
      int pos = -1;
      while (++pos < int(line.size() - 2)) {
        if (pos > 80 || (line[pos] == '/' && line[pos+1] == '/')) break;
        if (line[pos] == '\"') inQuotes = !inQuotes;
        if (inQuotes) continue;
        // Always preserve ->, ==, ++, --. 
        if ( (line[pos] == '-' && line[pos+1] == '>') ||
          (line[pos] == '-' && line[pos+1] == '-') ||  
          (line[pos] == '+' && line[pos+1] == '+') ) {
          ++pos;
          continue;
        }
        // Spaces before comparison operators. (After is added below.)
        if ( (line[pos] == '+' && line[pos+1] == '=')  ||     
          (line[pos] == '-' && line[pos+1] == '=')  ||     
          (line[pos] == '/' && line[pos+1] == '=')  ||     
          (line[pos] == '*' && line[pos+1] == '=')  ||     
          (line[pos] == '>' && line[pos+1] == '=')  ||          
          (line[pos] == '<' && line[pos+1] == '=')  ||          
          (line[pos] == '!' && line[pos+1] == '=')  ||          
          (line[pos] == '=' && line[pos+1] == '=') ) {
          if (pos >= 1 && line[pos-1] != ' ') {
            line.insert(pos," ");
            ++pos;
            ++nSpaces;
          }
          continue;
        }
        // Insert space before equals sign as well.
        if (pos >= 1 && line[pos+1] == '=' && line[pos] != ' ' ) {
          line.insert(++pos," ");
          ++nSpaces;
        }
        // Space after comma and =. 
        if ( (line[pos] == ',' || line[pos] == '=') && line[pos+1] != ' ') {
          line.insert(++pos," ");
          nSpaces++;
        }
        // Space around +.
        // (- still has to be done manually for the reasons above.)
        if ( line[pos] == '+' && line[pos+1] != ' ' ) {
          line.insert(++pos," ");
          nSpaces++;
        } else if ( line[pos+1] == '+' && line[pos] != ' '
          && line[pos+2] != '+') {          
          line.insert(++pos," ");
          nSpaces++;
        }
      }
    }
    
    // Detect long lines or lines with tab characters.
    if (warnLong && line.size() > maxLen) {
      cout << "    Warning: line " << i + 1 << " is " << line.size() 
           << " characters long" << endl;
      allFine = false;
    }
    if (line.find("\t") != string::npos) {
      cout << "    Warning: line " << i + 1 << " contains a tab character " 
           << endl;
      allFine = false;
    }

    // Replace processed line.
    lines[i] = line;
  }

  // Remove empty lines at the end of the file.
  if (doPad) {while (lines.back().length() == 1) lines.pop_back();}
  else       {while (lines.back().length() == 0) lines.pop_back();}
  nRemove -= lines.size();

  // Open output file.
  ofstream os( fileName.c_str() );
  if (!os) {
    cout << " Error: output file " << fileName << " not found" << endl;
    return false;
  }

  // Write out all lines.
  for (unsigned int i = 0; i < lines.size(); ++i) os << lines[i] << endl;
  cout << " Successfully processed " << lines.size() << " lines" << endl;
  if (nTrim > 0) cout << "    with " << nTrim << " line trims" << endl;
  if (nRemove > 0) cout << "    with " << nRemove << " line removals" << endl;
  if (nReplace > 0) cout << "    with " << nReplace << " replacements" << endl;
  if (nSpaces > 0) cout << "    with " << nSpaces << " spaces added" << endl;

  // Done.
  return allFine;
}

//==========================================================================
